from .strings import remove_suffix, ensure_suffix, count_up
from .code_rendering import render_and_write_code, render_template
from .numerical import random_direction, orthonormalise
from .collect_arguments import collect_arguments
from .input_handling import handle_input, sort_helpers, sympify_helpers

from sys import version_info

if version_info < (3,):
	from .module_handling27 import get_module_path, modulename_from_path, find_and_load_module, module_from_path
elif (3,) <= version_info < (3,3):
	raise NotImplementedError("Module loading for Python versions between 3 and 3.3 was not implemented. Please upgrade to a newer Python version.")
elif (3,3) <= version_info < (3,5):
	from .module_handling33 import get_module_path, modulename_from_path, find_and_load_module, module_from_path
elif (3,5) <= version_info:
	from .module_handling35 import get_module_path, modulename_from_path, find_and_load_module, module_from_path

try:
	from . import version
except ImportError:
	from warnings import warn
	warn('Failed to find (autogenerated) version.py. Do not worry about this unless you really need to know the version.')



from __future__ import print_function, division, with_statement
from os import path
from imp import find_module, load_module, load_dynamic, get_suffixes

raw_suffices = [S[0] for S in get_suffixes() if S[2]==3 and S[0][0]=="."]
suffices = sorted(raw_suffices, key=len)

def remove_suffix(path):
	for suffix in reversed(suffices):
		if path.endswith(suffix):
			return path.rpartition(suffix)[0]
	else:
		return path

def add_suffix(path):
	for suffix in suffices:
		if path.endswith(suffix):
			return path
	else:
		return path + suffices[0]

def modulename_from_path(full_path):
	filename = path.basename(full_path)
	return remove_suffix(filename)

def get_module_path(modulename, folder=""):
	return find_module(modulename, [folder])[1]

def find_and_load_module(modulename, folder=""):
	specs = find_module(modulename, [folder])
	return load_module(modulename, *specs)

def module_from_path(full_path):
	return load_dynamic( modulename_from_path(full_path), full_path )




#!/usr/bin/python
# -*- coding: utf-8 -*-

from jitcxde_common.input_handling import sort_helpers, filter_helpers, copy_helpers
import unittest
from sympy import symbols
from itertools import permutations
from sympy.abc import p,q,r,s,u,v

cycle = [ [p,q], [q,r], [r,p] ]
chain = [ [r,s], [q,r], [p,q] ]

class SortingTest(unittest.TestCase):
	def test_cyclic(self):
		for permutation in permutations(cycle):
			with self.assertRaises(ValueError):
				sort_helpers(permutation)
	
	def test_sorting(self):
		for permutation in permutations(chain):
			result = sort_helpers(list(permutation))
			assert result==chain

class FilterTest(unittest.TestCase):
	def test_all_needed(self):
		assert chain==filter_helpers(chain,{p})
	
	def test_spurious(self):
		chain_with_spurious = chain+[[u,v]]
		assert chain==filter_helpers(chain_with_spurious,{p})

class CopyTest(unittest.TestCase):
	def test_copy(self):
		assert chain==copy_helpers(chain)
	
	def test_independence(self):
		copy = copy_helpers(chain)
		copy.append([u,v])
		assert copy!=chain

if __name__ == "__main__":
	unittest.main(buffer=True)





from __future__ import print_function, division, with_statement
from sympy import preorder_traversal

def collect_arguments(expression, function):
	"""
	Parameters
	----------
	expression : SymPy expression
	function : Sympy function
	
	Returns
	-------
	arguments : list of SymPy expressions
		list of all arguments with which `function` is called within `expression`.
	"""
	
	arguments = set()
	for subexpression in preorder_traversal(expression):
		if subexpression.__class__ == function:
			arguments.add(subexpression.args)
	return arguments



from __future__ import print_function, division, with_statement
from inspect import isgeneratorfunction
from sympy import sympify

def prepare(expr):
	return sympify(expr).doit()

def handle_input(f_sym,n):
	"""
	Converts f_sym to a generator function, if necessary.
	Ensures that n is the length of f_sym, if not predefined.
	"""
	if isgeneratorfunction(f_sym):
		n = n or sum(1 for _ in f_sym())
		return (lambda: (prepare(entry) for entry in f_sym())), n
	elif len(f_sym) == 0:
		return lambda:f_sym, n
	else:
		len_f = len(f_sym)
		if (n is not None) and (len_f != n):
			raise ValueError("len(f_sym) and n do not match.")
		return (lambda: (prepare(entry) for entry in f_sym)), len_f

def depends_on_any(helper, other_helpers):
	for other_helper in other_helpers:
		if helper[1].has(other_helper[0]):
			return True
	return False

def sort_helpers(helpers):
	"""
	sorts a list of helpers such that no helper depends on a later one.
	"""
	
	if len(helpers)>1:
		for j,helper in enumerate(helpers):
			if not depends_on_any(helper, helpers):
				helpers.insert(0,helpers.pop(j))
				break
		else:
			raise ValueError("Helpers have cyclic dependencies.")
		
		helpers[1:] = sort_helpers(helpers[1:])
	
	return helpers

def sympify_helpers(helpers):
	return [(helper[0], sympify(helper[1]).doit()) for helper in helpers]

def filter_helpers(helpers,symbols):
	"""
	filters a list of helpers to contain only those listed in symbols and those needed for calculating them.
	"""
	needed = set(symbols) # convert and copy
	
	filtered_rev = []
	for helper in reversed(helpers):
		if helper[0] in needed:
			filtered_rev.append(helper)
			needed.update(helper[1].free_symbols)
	
	return list(reversed(filtered_rev))

def copy_helpers(helpers):
	return [helper for helper in helpers]




from __future__ import print_function

import sympy
import unittest
from jitcxde_common.collect_arguments import collect_arguments

class TestCollectArguments(unittest.TestCase):
	def test_complex_expression(self):
		f = sympy.Function("f")
		g = sympy.Function("g")
		a = sympy.Symbol("a")
		
		expression = 3**f(42) + 23 - f(43,44) + f(45+a)*sympy.sin( g(f(46,47,48)+17) - g(4) )
		
		self.assertEqual(
			collect_arguments(expression, f),
			{
				(sympy.Integer(42),),
				(sympy.Integer(43), sympy.Integer(44)),
				(45+a,),
				(sympy.Integer(46), sympy.Integer(47), sympy.Integer(48))}
			)

if __name__ == "__main__":
	unittest.main(buffer=True)



from sys import version_info

if version_info < (3,):
	from .module_handling27 import get_module_path, modulename_from_path, find_and_load_module, module_from_path, add_suffix
elif (3,) <= version_info < (3,3):
	raise NotImplementedError("Module loading for Python versions between 3 and 3.3 was not implemented. Please upgrade to a newer Python version.")
elif (3,3) <= version_info < (3,5):
	from .module_handling33 import get_module_path, modulename_from_path, find_and_load_module, module_from_path, add_suffix
elif (3,5) <= version_info:
	from .module_handling35 import get_module_path, modulename_from_path, find_and_load_module, module_from_path, add_suffix




from __future__ import print_function, division, with_statement

# String manipulation
# -------------------

# obsolete
def remove_suffix(string, suffix):
	partition = string.rpartition(suffix)
	if partition[1] and not partition[2]:
		return partition[0]
	else:
		return string

# obsolete
def ensure_suffix(string, suffix):
	if not string.endswith(suffix):
		return string + suffix
	else:
		return string

def rsplit_int(s):
	if s and s[-1].isdigit():
		x,y = rsplit_int(s[:-1])
		return x, y+s[-1]
	else:
		return s, ""

def count_up(name):
	s, i = rsplit_int(name)
	return s + ( "%%.%ii" % len(i) % (int(i)+1)  if i else "_1" )




from __future__ import print_function, division, with_statement
from sys import stderr
from os import path
from inspect import stack
from warnings import warn
from itertools import chain
from sympy.core.cache import clear_cache
from sympy import __version__ as sympy_version
from sympy.printing.ccode import ccode
from jinja2 import Environment, FileSystemLoader
from jitcxde_common.strings import count_up

def check_code(code):
	if code.startswith("// Not"):
		stderr.write(code)
		raise Exception("The above expression could not be converted to C Code.")
	return code

def render_declarator(name, _type, size=0):
	return _type + " " + name + ("[%i]"%size if size else "")
	
def write_in_chunks(lines, mainfile, deffile, name, chunk_size, arguments):
	funcname = "definitions_" + name
	
	first_chunk = []
	try:
		for _ in range(chunk_size+1):
			first_chunk.append(next(lines))
	except StopIteration:
		for line in first_chunk:
			mainfile.write(line)
	else:
		lines = chain(first_chunk, lines)
		
		if sympy_version >= "1":
			clear_sympy_cache = clear_cache
		else:
			def clear_sympy_cache(): pass
			warn("Not clearing SymPy cache between chunks because this is buggy in this SymPy version. If excessive memory is used, this is why and you have to upgrade SymPy.")
		
		while True:
			mainfile.write(funcname + "(")
			deffile.write("void " + funcname + "(")
			if arguments:
				mainfile.write(", ".join(argument[0] for argument in arguments))
				deffile.write(", ".join(render_declarator(*argument) for argument in arguments))
			else:
				deffile.write("void")
			mainfile.write(");\n")
			deffile.write("){\n")
			
			try:
				for _ in range(chunk_size):
					deffile.write(next(lines))
			except StopIteration:
				break
			finally:
				deffile.write("}\n")
			
			funcname = count_up(funcname)
			clear_sympy_cache()

def render_and_write_code(
		expressions,
		tmpfile,
		name,
		functions = (),
		chunk_size = 100,
		arguments = ()
		):
	
	user_functions = {function:function for function in functions}
	
	def codelines():
		for expression in expressions:
			codeline = ccode(expression, user_functions=user_functions)
			yield check_code(codeline) + ";\n"
	
	with \
			open( tmpfile(name+".c"            ), "w" ) as mainfile, \
			open( tmpfile(name+"_definitions.c"), "w" ) as deffile:
		if chunk_size < 1:
			for line in codelines():
				mainfile.write(line)
		else:
			write_in_chunks(codelines(), mainfile, deffile, name, chunk_size, arguments)

def render_template(filename, target, folder=None, **kwargs):
	folder = folder or path.dirname( stack()[1][1] )
	env = Environment(loader=FileSystemLoader(folder))
	template = env.get_template(filename)
	with open(target, "w") as codefile:
		codefile.write(template.render(kwargs))




from .strings import remove_suffix, ensure_suffix, count_up
from .code_rendering import render_and_write_code, render_template
from .numerical import random_direction, orthonormalise, rel_dist
from .collect_arguments import collect_arguments
from .input_handling import handle_input, sort_helpers, sympify_helpers, copy_helpers, filter_helpers
from .module_handling import get_module_path, modulename_from_path, find_and_load_module, module_from_path, add_suffix
from ._jitcxde import jitcxde, DEFAULT_COMPILE_ARGS, DEFAULT_LINK_ARGS, MSVC_COMPILE_ARGS,MSVC_LINK_ARGS

try:
	from .version import version as __version__
except ImportError:
	from warnings import warn
	warn('Failed to find (autogenerated) version.py. Do not worry about this unless you really need to know the version.')



#!/usr/bin/python3
# -*- coding: utf-8 -*-

from __future__ import print_function

from tempfile import mkdtemp
from os import path
from inspect import stack
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
import shutil
from sys import version_info, modules
from warnings import warn
from traceback import format_exc

import numpy
from jinja2 import Environment, FileSystemLoader

from .module_handling import get_module_path, modulename_from_path, find_and_load_module, module_from_path, add_suffix
from .strings import count_up

#: A list with the default extra compile arguments. Note that without `-Ofast`, `-ffast-math`, or `-funsafe-math-optimizations` (if supported by your compiler), you may experience a considerable speed loss since SymPy uses the `pow` function for small integer powers (`SymPy Issue 8997`_).
DEFAULT_COMPILE_ARGS = [
			"-std=c11",
			"-Ofast",
			"-g0",
			"-march=native",
			"-mtune=native",
			"-Wno-unknown-pragmas",
			]

#: A list with the default linker arguments.
DEFAULT_LINK_ARGS = [ "-lm" ]

#: A list with the default compile arguments for the Microsoft compiler. I could not find what level of optimisation is needed to address the problem of SymPy using the `pow` function for small integer powers (`SymPy Issue 8997`_).
MSVC_COMPILE_ARGS = [
			"/Ox",
			"/wd4068",
			"/wd4146",
			"/wd4018"
			]

#: A list with the default linker arguments for the Microsoft compiler.
MSVC_LINK_ARGS = [ "/ignore:4197" ]

class build_ext_with_compiler_detection(build_ext):
	def build_extensions(self):
		for extension in self.extensions:
			if self.compiler.compiler_type=="msvc":
				extension.extra_compile_args = MSVC_COMPILE_ARGS
				extension.extra_link_args    = MSVC_LINK_ARGS
			else:
				extension.extra_compile_args = DEFAULT_COMPILE_ARGS
				extension.extra_link_args    = DEFAULT_LINK_ARGS
		
		build_ext.build_extensions(self)

class jitcxde(object):
	"""
	A base class containing elementary, common functionalities of all JiTC*DE projects – mostly file handling. It is pretty dysfunctional on its own and only made to be inherited from.
	"""
	
	def __init__(self,verbose,module_location):
		self._tmpdir = None
		self.verbose = verbose
		self._modulename = "jitced"
		
		if module_location is not None:
			self.jitced = module_from_path(module_location)
			self.compile_attempt = True
		else:
			self.jitced = None
			self.compile_attempt = None
		
	
	def _tmpfile(self,filename=None):
		if self._tmpdir is None:
			self._tmpdir = mkdtemp()
		
		if filename is None:
			return self._tmpdir
		else:
			return path.join(self._tmpdir, filename)
	
	def report(self,message):
		if self.verbose:
			print(message)
	
	def _process_modulename(self,modulename):
		"""
		Sets the modulename from input (if specified) or automatically.
		"""
		if modulename:
			if modulename in modules.keys():
				raise NameError("Module name has already been used in this instance of Python.")
			self._modulename = modulename
		else:
			while self._modulename in modules.keys():
				self._modulename = count_up(self._modulename)
		
		modulefile = self._tmpfile(self._modulename + ".so")
		if path.isfile(modulefile):
			raise OSError("Module file already exists.")
	
	@property
	def sourcefile(self):
		return self._tmpfile(self._modulename + ".c")
	
	def _render_template(self,**kwargs):
		kwargs["module_name"] = self._modulename
		kwargs["Python_version"] = version_info[0]
		folder = path.dirname( stack()[1][1] )
		env = Environment(loader=FileSystemLoader(folder))
		template = env.get_template("jitced_template.c")
		with open(self.sourcefile, "w") as codefile:
			codefile.write(template.render(kwargs))
	
	def _attempt_compilation(self,reset=True):
		self.report("Generating, compiling, and loading C code.")
		try:
			self.compile_C()
		except:
			warn(format_exc())
			line = "\n"+60*"="+"\n"
			warn(line + "READ ME FIRST" + line + "Generating compiled integrator failed; resorting to lambdified functions. If you can live with using the Python backend, you can call generate_lambdas to explicitly do this and bypass the compile attempt and error messages. Otherwise, you want to take care of fixing the above errors." + 2*line)
		else:
			if reset:
				self.reset_integrator()
	
	def _compile_and_load(self,verbose,extra_compile_args,extra_link_args=None):
		auto_args = extra_link_args is None and extra_compile_args is None
		
		extension = Extension(
				self._modulename,
				sources = [self.sourcefile],
				extra_link_args = extra_link_args,
				include_dirs = [numpy.get_include()],
				extra_compile_args = extra_compile_args,
				)
		
		script_args = [
				"build_ext",
				"--build-lib", self._tmpfile(),
				"--build-temp", self._tmpfile(),
				"--force",
				#"clean" #, "--all"
				]
		
		setup(
			name = self._modulename,
			ext_modules = [extension],
			script_args = script_args,
			verbose = verbose,
			cmdclass = {'build_ext':build_ext_with_compiler_detection} if auto_args else {}
			)
		
		self.jitced = find_and_load_module(self._modulename,self._tmpfile())
		self.compile_attempt = True
	
	def save_compiled(self, destination="", overwrite=False):
		"""
		saves the module file with the compiled functions for later use (see the `module_location` argument). If no compiled derivative exists, it tries to compile it first using `compile_C`. In most circumstances, you should not rename this file, as the filename is needed to determine the module name.
		
		Parameters
		----------
		destination : string specifying a path
			If this specifies only a directory (don’t forget the trailing `/` or similar), the module will be saved to that directory. If empty (default), the module will be saved to the current working directory. Otherwise, the functions will be (re)compiled to match that filename. A file ending will be appended if needed.
		overwrite : boolean
			Whether to overwrite the specified target if it already exists.
		
		Returns
		-------
		filename : string
			The destination that was actually used.
		"""
		
		folder, filename = path.split(destination)
		
		if filename:
			destination = add_suffix(destination)
			modulename = modulename_from_path(filename)
			if modulename != self._modulename:
				self.compile_C(modulename=modulename)
				self.report("compiled C code")
			else:
				self._compile_C()
			sourcefile = get_module_path(self._modulename, self._tmpfile())
		else:
			self._compile_C()
			sourcefile = get_module_path(self._modulename, self._tmpfile())
			destination = path.join(folder, add_suffix(self._modulename))
			self.report("saving file to " + destination)
		
		if not self.compile_attempt:
			raise RuntimeError("Compilation failed. Cannot save module file.")
		
		if path.isfile(destination) and not overwrite:
			raise OSError("Target File already exists and \"overwrite\" is set to False")
		else:
			shutil.copy(sourcefile, destination)
		
		return destination
	
	def __del__(self):
		try:
			shutil.rmtree(self._tmpdir)
		except (OSError, AttributeError, TypeError):
			pass

